#                                testutil.tcl
#
# Utility procedures for TCL test packages.  These procedures all expect a 
# global "ModuleName" to be define, which contains the name of the test package
# being run.  This may be a short string.

# Make sure invalid commands are trapped.

set invalid_command_name_hook ""


# Check that a value is equal to what was expected.  If not, generate
# an error

proc check {got expect idnum} {
  global ModuleName
  if {[string compare $got $expect] != 0} {
    error [format {%s test error %s: wanted "%s", got "%s"} \
	           $ModuleName $idnum $expect $got]}
}


# Check that a numeric value is equal to what was expected.  If not, generate
# an error

proc checknum {got expect idnum} {
  global ModuleName
  if {$got != $expect} {
    error [format {%s test error %s: wanted "%s", got "%s"} \
	           $ModuleName $idnum $expect $got]}
}


# Run and verify a command that should return TCL_OK. Returns the value
# returned by the command, if that needs to be verified.

proc do0cmd {cmd idnum} {
  global ModuleName
  if {[catch {set RetVal [uplevel $cmd]} msg] != 0} {
    error [format {%s test error %s: Unexpected error: %s: %s} \
                   $ModuleName $idnum $cmd $msg]}
  return $RetVal
}

# Run and verify a command that should return TCL_ERROR, pass back the returned
# error message

proc do1cmd {cmd msgvar idnum} {
  global ModuleName
  if {[catch {uplevel $cmd} msg] != 1} {
    error [format {%s test error %s: No error returned: %s} \
          $ModuleName $idnum $cmd]}
  uplevel "set $msgvar \{$msg\}"
}


# Increment a name.  This takes a name and "adds one" to it, that is advancing
# each digit lexically through "0"..."9" -> "A"-"Z" -> "a"..."z".  When one
# digit wraps, the next one is advanced.

proc IncrName {Name} {
    set Last  [expr [clength $Name]-1]
    set Begin [substr $Name 0 $Last]
    set Digit [cindex $Name $Last]
    case $Digit in \
        {9}     {set Digit A} \
        {Z}     {set Digit a} \
        {z}     {if {$Last == 0} then {
                     return 0 ;# Wrap around
                 } else {
                     return [IncrName $Begin]0}} \
        default {set Digit [char [expr [ordinal $Digit]+1]]}
    return $Begin$Digit
}


# If the strequ command is not defined (as with a simple baseline), create
# a proc to handle it.

if {[string compare [info command strequ] ""] == 0} {
    proc strequ {str1 str2} {return [expr [string compare $str1 $str2]==0]}
}

        
