/* 
 * cmdloop --
 *
 * based on code:
 *
 * Copyright 1987 Regents of the University of California
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "%M% %I% %H%";
#endif

#include <stdio.h>
#include <errno.h>

#include "tcl.h"

#if defined(XENIX) || defined(BSD)
extern int errno;
#endif

extern int signal_received;


/*
 *----------------------------------------------------------------------
 *
 * print_result --
 *
 *      Print a Tcl result
 *
 * Results:
 *
 *      Takes an open file pointer, a return value and some result
 *      text.  Prints the result text if the return value is TCL_OK,
 *      prints "Error:" and the result text if it's TCL_ERROR,
 *      else prints "Bad return code:" and the result text.
 *
 *----------------------------------------------------------------------
 */
void
print_result(fp, returnval, result_text)
    FILE   *fp;
    int     returnval;
    char   *result_text;
{
    if (returnval == TCL_OK)
        {
        if (result_text && *result_text != 0)
            fprintf(fp, "%s\n", result_text);
        }
     else
        {
        fprintf(stderr, "%s: %s\n",
                (returnval == TCL_ERROR) ? "Error" : "Bad return code",
                result_text);
        fflush (fp);
        fflush (stderr);
        }
}

/*
 *----------------------------------------------------------------------
 *
 * OutputPrompt --
 *     Outputs a prompt by executing either the command string in
 *     toplevel_prompt_hook or downlevel_prompt_hook.
 *
 *----------------------------------------------------------------------
 */
static void
OutputPrompt (interp, outFP, topLevel)
    Tcl_Interp *interp;
    FILE       *outFP;
    int         topLevel;
{
    char *hookName;
    char *promptHook;
    int   result;
    int   promptDone = FALSE;

    hookName = topLevel ? "toplevel_prompt_hook"
                        : "downlevel_prompt_hook";
    if (((promptHook = Tcl_GetVar (interp, hookName, 1)) != NULL) &&
         (*promptHook != '\0')) 
        {
        result = Tcl_Eval(interp, promptHook, 0, (char **)NULL);
        if (!((result == TCL_OK) || (result == TCL_RETURN)))
            {
            fprintf(outFP, "Error in prompt hook: %s", 
                    interp->result);
            print_result(outFP, result, interp->result);
            }
        else
            {
            fputs (interp->result, outFP);
            promptDone = TRUE;
            }
        } 
    if (!promptDone)
        {
        if (topLevel)
            fputs("%", outFP);
        else
            fputs(">", outFP);
        }
    fflush(outFP);

}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CommandLoop --
 *
 *      Run a Tcl command loop
 *
 * Results:
 *
 *      Takes an interpreter, in and out file handles and an
 *      interactive flag and reads and executes everything
 *      it reads from input.
 *
 * Side effects:
 *
 *      Anything that happens as a result of user or file input.
 *
 *----------------------------------------------------------------------
 */
Tcl_CommandLoop (interp, in, out, interactive)
    Tcl_Interp *interp;
    FILE       *in;
    FILE       *out;
    int         interactive;
{
    Tcl_CmdBuf cmdBuf;
    char       s[256];
    int        command_complete = 1;
    int        result;
    char      *cmd;

    cmdBuf = Tcl_CreateCmdBuf();

    while (1)
        {

        clearerr(in);
        clearerr(out);
        OutputPrompt (interp, out, command_complete);
        errno = 0;
        if (fgets(s, sizeof(s), in) == NULL)
            {
            if (!feof(in) && interactive && (errno == EINTR))
                {
                signal_received = 0;
                putchar('\n');
                continue;  /* Go get the next command */
                }
            goto endOfFile;
            }
        cmd = Tcl_AssembleCmd(cmdBuf, s);

        if (cmd == NULL)
            command_complete = 0;
        else
            {
#ifdef NOHISTORY
            result = Tcl_Eval(interp, cmd, 0, (char **)NULL);
#else
            result = Tcl_RecordAndEval(interp, cmd, 0);
#endif
            if (interactive || (result != TCL_OK))
                print_result(out, result, interp->result);

            command_complete = 1;
            }
        }
endOfFile:
    Tcl_DeleteCmdBuf(cmdBuf);
}

/*
 *----------------------------------------------------------------------
 *
 * SetPromptVar --
 *     Set one of the prompt hook variables, saving a copy of the old
 *     value, if it exists.
 *
 *----------------------------------------------------------------------
 */
static char*
SetPromptVar (interp, hookVarName, newHookValue)
    Tcl_Interp *interp;
    char       *hookVarName;
    char       *newHookValue;
{
    char *hookPtr;    
    char *oldHookPtr = NULL;

    hookPtr = Tcl_GetVar (interp, hookVarName, TRUE);
    if (hookPtr != NULL)
        {
        oldHookPtr = ckalloc (strlen (hookPtr) + 1);
        strcpy (oldHookPtr, hookPtr);
        }
    Tcl_SetVar (interp, hookVarName, newHookValue, TRUE);

    return oldHookPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * ResetPromptVar --
 *     Reset one of the prompt hook variables to its old value and return
 *     the storage used to hold the old value.
 *     This should release the variable if it did not exists???
 *
 *----------------------------------------------------------------------
 */
static void
ResetPromptVar (interp, hookVarName, oldHookValue)
    Tcl_Interp *interp;
    char       *hookVarName;
    char       *oldHookValue;
{
    char *hookValue;    
 
    hookValue = (oldHookValue != NULL) ? oldHookValue : "";
    Tcl_SetVar (interp, hookVarName, hookValue, TRUE);
    
    if (oldHookValue != NULL)
        ckfree (oldHookValue);
    
}

/*
 *----------------------------------------------------------------------
 *
 * CommandloopCmd --
 *     Implements the TCL commandloop command:
 *       commandloop prompt prompt2
 *     See the cmdloop(TCL) manual page.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CommandloopCmd(clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    char *oldTopLevelHook  = NULL;
    char *oldDownLevelHook = NULL;

    if (argc > 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s [prompt] [prompt2]", argv[0]);
    if (argc > 1)
        oldTopLevelHook = SetPromptVar (interp, "toplevel_prompt_hook",
                                        argv[1]);
    if (argc > 2)
        oldDownLevelHook = SetPromptVar (interp, "downlevel_prompt_hook",
                                         argv[2]);

    Tcl_CommandLoop (interp, stdin, stdout, TRUE);

    if (argc > 1)
        ResetPromptVar (interp, "toplevel_prompt_hook", oldTopLevelHook);
    if (argc > 2)
        ResetPromptVar (interp, "downlevel_prompt_hook", oldDownLevelHook);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitCommandLoop --
 *   Initialize commandloop command and install it into Tcl
 *
 *----------------------------------------------------------------------
 */
Tcl_InitCommandLoop(interp)
    Tcl_Interp *interp;
{
Tcl_CreateCommand(interp, "commandloop", CommandloopCmd, (ClientData)NULL, 
                  (void (*)())NULL);

}

