/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
    static char *sccs_id = "%M% %I% %H%";
#endif

#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#if defined(XENIX) || defined(BSD)
#  define CHAR_MAX 255
#else
#  include <limits.h>
#endif

#if XENIX286
  typedef int size_t;
#endif

#include "tcl.h"
#include "iocmds.h"
#include "regexp.h"

extern struct handle handles[MAXFILES];

/*
 * A scan context describes a collection of match patterns and commands,
 * along with a match default command to apply to a file on a scan.
 */
 
#define CONTEXT_IN_USE_FLAG 1
#define CONTEXT_A_CASE_INSENSITIVE_FLAG 2
#define MATCH_CASE_INSENSITIVE_FLAG 4

typedef struct matchDef_t {
    regexp_t            regExpInfo;
    char               *command;
    struct matchDef_t  *nextMatchDefPtr;
    short               matchflags;
    } matchDef_t;
typedef struct matchDef_t *matchDef_pt;

typedef struct scanContext_t {
    matchDef_pt  matchListHead;
    matchDef_pt  matchListTail;
    char        *defaultAction;
    short        flags;
    } scanContext_t;
typedef struct scanContext_t *scanContext_pt;

#define MAX_SCANCONTEXTS 32

scanContext_t scancontexts [MAX_SCANCONTEXTS];
int current_scan_context = 0;


/*
 *----------------------------------------------------------------------
 *
 * IsContextOpen --
 *    Check is a scan context is open, and if it is not, return a TCL
 *    error in the result.
 *
 * Results:
 *    TRUE or FALSE.
 *
 *----------------------------------------------------------------------
 */
static int
IsContextOpen (interp, contextIdx)
    Tcl_Interp  *interp;
    int          contextIdx;
{
    if ((scancontexts[contextIdx].flags & CONTEXT_IN_USE_FLAG) == 0)
        {
        Tcl_RetError (interp, "TCL", "scan context not open (context%d)", 
                      contextIdx);
        return FALSE;
        }
    return TRUE;
}

/*
 *----------------------------------------------------------------------
 *
 * convcontext --
 *    Convert a scan context name to an index into the context table.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
static int
convcontext (interp, handle, index)
    Tcl_Interp *interp;
    char       *handle;
    int        *index;
{
    int result;

    if ((strncmp(handle, "context", 7) != 0) ||
        (!StrToInt (&handle[7], 10, &result))||
        (result < 0) ||
        (result >= MAX_SCANCONTEXTS))
            return Tcl_RetError (interp, "TCL", "Invalid scan context (%s)", 
                                 handle);

    if (!IsContextOpen (interp, result))
        return TCL_ERROR;

    *index = result;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdCreate_scancontext --
 *     Implements the TCL CmdCreate_scancontext command:
 *         create_scancontext
 *     See the filescan(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 * Side effects:
 *      Sets the current scan context.
 *
 *----------------------------------------------------------------------
 */
int
CmdCreate_scancontext (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int i;
for (i = 0; i < MAX_SCANCONTEXTS; i++) {
    if ((scancontexts[i].flags & CONTEXT_IN_USE_FLAG) == 0) {
        scancontexts[i].flags = CONTEXT_IN_USE_FLAG; 
        scancontexts[i].matchListHead = NULL;
        scancontexts[i].matchListTail = NULL;
        scancontexts[i].defaultAction = NULL;
        current_scan_context = i;
        sprintf(interp->result, "context%d", i);
        return TCL_OK;
    }
}
}

int
CmdCurrent_scancontext (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        sprintf(interp->result, "context%d", current_scan_context);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSelect_scancontext --
 *     Implements the TCL CmdSelect_scancontext command:
 *         select_scancontext scancontext
 *     See the filescan(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdSelect_scancontext (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int index;

if (argc != 2)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s contexthandle", 
                         argv[0]);

if (convcontext (interp, argv[1], &index) != TCL_OK)
    return TCL_ERROR;

current_scan_context = index;
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdDelete_scancontext --
 *     Implements the TCL CmdDelete_scancontext command:
 *         delete_scancontext scancontext
 *     See the filescan(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdDelete_scancontext (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int            index;
scanContext_pt cpr;
matchDef_pt    opr, mpr;

if (argc != 2)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s contexthandle",
                         argv[0]);

if (convcontext(interp, argv[1], &index) != TCL_OK)
    return TCL_ERROR;

cpr = &scancontexts[index];
for (mpr = cpr->matchListHead; mpr != NULL;) {
    CleanRegExpInfo (&mpr->regExpInfo);
    if (mpr->command != NULL)
        ckfree(mpr->command);
    opr = mpr;
    mpr = mpr->nextMatchDefPtr;
    ckfree(opr);
    }
cpr->matchListHead = NULL;
cpr->matchListTail = NULL;

if (cpr->defaultAction != NULL) {
    ckfree(cpr->defaultAction);
    cpr->defaultAction = NULL;
    }
cpr->flags = 0;
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdMatch --
 *     Implements the TCL match command:
 *         match pattern [nocase] command
 *     See the filescan(TCL) manual page.  This uses both Boyer_Moore 
 *     and regular expressions matching.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdMatch (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
scanContext_pt  cpr;
char           *result;
matchDef_pt     newmatch;
int             compFlags = BOTH_ALGORITHMS;

if (!IsContextOpen (interp, current_scan_context))
    return TCL_ERROR;
cpr = &scancontexts [current_scan_context];

if ((argc < 3) || (argc > 4) ||
    ((argc == 4) && (strcmp(argv[2], "nocase") != 0)))
    return Tcl_RetError (interp, "TCL", 
           "bad # args: %.50s regular_expression [nocase] command", argv[0]);

newmatch = (matchDef_pt)ckalloc(sizeof(matchDef_t));
newmatch->matchflags = 0;
InitRegExpInfo (&newmatch->regExpInfo);
if (compFlags & NO_CASE) {
    newmatch->matchflags |= MATCH_CASE_INSENSITIVE_FLAG;
    cpr->flags |= CONTEXT_A_CASE_INSENSITIVE_FLAG;
}

if (RegExpCompile (interp, &newmatch->regExpInfo, argv [1], compFlags) != 
    TCL_OK) {
    ckfree (newmatch);
    return (TCL_ERROR);
}

newmatch->command = ckalloc(strlen(argv[argc-1]) + 1);
strcpy(newmatch->command, argv[argc-1]);

/*
 * Link in the new match.
 */
newmatch->nextMatchDefPtr = NULL;
if (cpr->matchListHead == NULL)
    cpr->matchListHead = newmatch;
else
    cpr->matchListTail->nextMatchDefPtr = newmatch;
cpr->matchListTail = newmatch;

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdMatch_default --
 *     Implements the TCL match_default command:
 *         match_default command
 *     See the filescan(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdMatch_default (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
scanContext_pt cpr;
matchDef_pt    newmatch;

if (!IsContextOpen (interp, current_scan_context))
    return TCL_ERROR;
cpr = &scancontexts[current_scan_context];

if (argc != 2)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s tcl_command",
                         argv[0]);

cpr->defaultAction = ckalloc(strlen(argv[1]) + 1);
strcpy(cpr->defaultAction, argv[1]);

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SetMatchVar --
 *     Sets the TCL array variable matchInfo to contain information 
 *     about the line that is matched.
 *
 * Side effects:
 *     A TCL array variable is created or altered.
 *
 *----------------------------------------------------------------------
 */
static void
SetMatchVar (interp, fileLine, fileOffset, scanLineNum, fileHandle)
    Tcl_Interp *interp;
    char       *fileLine;
    long        fileOffset;
    long        scanLineNum;
    char       *fileHandle;
{
char numBuf [20];

Tcl_SetArrayVar (interp, "matchInfo", "line", fileLine, 1);

sprintf (numBuf, "%ld", fileOffset);
Tcl_SetArrayVar (interp, "matchInfo", "offset", numBuf, 1);

sprintf (numBuf, "%ld", scanLineNum);
Tcl_SetArrayVar (interp, "matchInfo", "linenum", numBuf, 1);

Tcl_SetArrayVar (interp, "matchInfo", "handle", fileHandle, 1);
}

/*
 *----------------------------------------------------------------------
 *
 * CmdScan_file --
 *     Implements the TCL fscan command:
 *         scan_file filehandle [scancontext]
 *     See the filescan(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdScan_file (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
char            buf [1024];
char            lowercase_buf [1024];
FILE           *file;
int             filehandle;
matchDef_pt     mp;
int             result;
scanContext_pt  cpr;
int             matched_at_least_one;
long            fileOffset;
long            matchOffset;
long            scanLineNum = 0;
int             scan_context_index;

if ((argc != 2) && (argc != 3))
    return Tcl_RetError (interp, "TCL", 
                         "bad # arg: %.50s filehandle [scancontext]", argv[0]);

if (argc == 2) 
    {
    if (!IsContextOpen (interp, current_scan_context))
        return TCL_ERROR;
    scan_context_index = current_scan_context;
    }
else
    if (convcontext(interp, argv[2], &scan_context_index) != TCL_OK)
            return TCL_ERROR;

cpr = &scancontexts[scan_context_index];

if (cpr->matchListHead == NULL)
    return Tcl_RetError (interp, "TCL", "no patterns in current scan context");

if (convfile(interp, argv[0], argv[1], &filehandle, 1) != TCL_OK)
        return(TCL_ERROR);
file = handles[filehandle].fp;

result = TCL_OK;  /* Assume the best */

fileOffset = ftell (file);  /* Get starting offset */

while (result == TCL_OK)
    {
    int stored_this_line = FALSE;
    int nchars;

    if (fgets(buf, sizeof(buf), file) == NULL)
        break;

    scanLineNum++;
    nchars = strlen(buf) - 1;
    matchOffset = fileOffset;
    fileOffset += nchars + 1;
    buf[nchars] = '\0';
    stored_this_line = 0;
    matched_at_least_one = 0;
    if (cpr->flags & CONTEXT_A_CASE_INSENSITIVE_FLAG)
        DownShift (lowercase_buf, buf);

    for (mp = cpr->matchListHead; mp != NULL; mp = mp->nextMatchDefPtr) {

         if (!RegExpExecute (interp, &mp->regExpInfo, buf, lowercase_buf))
                continue;

         matched_at_least_one = TRUE;
         if (!stored_this_line) {
             SetMatchVar (interp, buf, matchOffset, scanLineNum, argv[1]);
             stored_this_line = TRUE;
         }

         result = Tcl_Eval(interp, mp->command, 0, (char **)NULL);
         if (result == TCL_CONTINUE) {
             result = TCL_OK;
             break;   /* Don't process any more matches for this line */
         }
         if (result == TCL_ERROR)
             Tcl_AddErrorInfo (interp, "while executing a match command");
         if (result != TCL_OK)
             break;

         }

        /* process default action if required */
        if (!matched_at_least_one && (cpr->defaultAction != NULL)) {
            if (!stored_this_line)
                SetMatchVar (interp, buf, matchOffset, scanLineNum, argv[1]);

            result = Tcl_Eval (interp, cpr->defaultAction, 0, (char **)NULL);
            if (result == TCL_CONTINUE)
                result = TCL_OK;    /* This doesn't mean anything, but don't */
                                    /* break the user.                      */
            if (result == TCL_ERROR)
             Tcl_AddErrorInfo (interp, 
                               "while executing a match_default command");
        }
    }
if (result == TCL_RETURN)
    result = TCL_OK;
return result;
}

/*
 *----------------------------------------------------------------------
 *
 *  Tcl_InitFilescan --
 *      Initialize the TCL file scanning facility..
 *
 *----------------------------------------------------------------------
 */
void
Tcl_InitFilescan (interp)
Tcl_Interp *interp;
{
int            idx;
scanContext_pt cpr;

for (idx = 0; idx < MAX_SCANCONTEXTS; idx++)
    {
    cpr = &scancontexts[idx];
    cpr->matchListHead = NULL;
    cpr->matchListTail = NULL;
    cpr->flags = 0;
    cpr->defaultAction = NULL;
    }

/* context 0 is defined by default */
scancontexts[0].flags = CONTEXT_IN_USE_FLAG;

Tcl_CreateCommand (interp, "scan_file", CmdScan_file, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "match", CmdMatch, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "match_default", CmdMatch_default,
                  (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand (interp, "create_scancontext", CmdCreate_scancontext,
                  (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand (interp, "select_scancontext", CmdSelect_scancontext,
                  (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand (interp, "delete_scancontext", CmdDelete_scancontext,
                  (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand (interp, "current_scancontext", CmdCurrent_scancontext,
                  (ClientData)NULL, (void (*)())NULL);
}

