/* 
 * string.c --
 *
 *      TCL extend string commands.
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "@(#) string.c 2.2 10/2/90";
#endif

#include <sys/types.h>
#include <ctype.h>
#include <string.h>

#include "tcl.h"

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CharCmd --
 *     Implements the char TCL command:
 *         char number
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Converts the numeric value, number to an ascii character. Number
 *      must be in the range 0 through 255.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_CharCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_CharCmd */
int number;

if (argc != 2)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s number", argv[0]);

if ((!StrToInt (argv[1], 10, &number)) || (number < 0) || (number > 255)) 
  return Tcl_RetError (interp, "TCL", "bad number \"%.50s\"", argv[1]);

interp->result [0] = number;
interp->result [1] = 0;
return TCL_OK;

} /* Tcl_CharCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CindexCmd --
 *     Implements the cindex TCL command:
 *         cindex string index
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns the character indexed by  index  (zero  based)  from
 *      string. 
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_CindexCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_CindexCmd */
int index;

if (argc != 3)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string index",
                       argv[0]);

if ((!StrToInt (argv[2], 10, &index)) || (index < 0)) 
  return Tcl_RetError (interp, "TCL", "bad index \"%.50s\"", argv[2]);

if (index >= strlen (argv[1]))
  return Tcl_RetError (interp, "TCL", "index out of range \"%.50s\"", argv[2]);

interp->result [0] = argv[1][index];
interp->result [1] = 0;
return TCL_OK;

} /* Tcl_CindexCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ClengthCmd --
 *     Implements the clength TCL command:
 *         clength string
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns the length of string in characters. 
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_ClengthCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_ClengthCmd */

if (argc != 2)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string", argv[0]);

sprintf (interp->result, "%d", strlen (argv[1]));
return TCL_OK;

} /* Tcl_ClengthCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CrangeCmd --
 *     Implements the crange and substr TCL commands:
 *         crange string first length
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Standard Tcl result.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_CrangeCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_CrangeCmd */
int fullLen;
int index;
int subLen;
int first;

if (argc != 4)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string first %s", 
                       argv[0], (*argv[0] == 'c') ? "last" : "length");

if ((!StrToInt (argv[2], 10, &first)) || (first < 0))
  return Tcl_RetError (interp->result, "TCL", "bad first \"%.50s\"", argv[2]);

fullLen = strlen (argv[1]);
if (first >= fullLen)
  return Tcl_RetError (interp, "TCL", "first out of range \"%.50s\"", argv[2]);

if (strcmp (argv[3], "end") == 0) {
  subLen = fullLen - first;}
else
  { /* Length or last specified */
  if (!StrToInt (argv[3], 10, &subLen) || (subLen < 0)) 
    return Tcl_RetError (interp, "TCL", "bad %s \"%.50s\"",
                         (*argv[0] == 'c') ? "last" : "length", argv[3]);
  if (*argv[0] == 'c') 
    { /* Last specified */
    if (subLen < first)
      return Tcl_RetError (interp, "TCL", "last is before first");
    subLen = subLen - first +1;
    } /* Last specified */ 
    
  if (first + subLen > fullLen)
    subLen = fullLen - first;

  } /* Length or last specified */

if (subLen >= TCL_RESULT_SIZE)
  { /* Must return a dynamic string */
  interp->result = (char *) ckalloc((unsigned) subLen + 1);
  interp->dynamic = TCL_DYNAMIC;
  } /* Must return a dynamic string */

strncpy (interp->result, argv[1] + first, subLen);
interp->result [subLen] = 0;
return TCL_OK;

} /* Tcl_CrangeCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ChangeCaseCmd --
 *     Implements the lowercase and uppercase TCL command:
 *         lowercase string
 *         uppercase string
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns a string mapped either to upper or lower case.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_ChangeCaseCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_ChangeCaseCmd */
register char *oldPtr;
register int   len;

if (argc != 2)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string", argv[0]);

oldPtr = argv[1];
len = strlen (oldPtr);
if (len >= TCL_RESULT_SIZE)
  { /* Must return a dynamic string */
  interp->result = (char *) ckalloc((unsigned) len + 1);
  interp->dynamic = TCL_DYNAMIC;
  } /* Must return a dynamic string */

if (*argv[0] == 'l')
  DownShift (interp->result, oldPtr);
else
  UpShift (interp->result, oldPtr);

return TCL_OK;

} /* Tcl_ChangeCaseCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OrdinalCmd --
 *     Implements the ordinal TCL command:
 *         ordinal string
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns the ordinal value of the first character in string.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_OrdinalCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_OrdinalCmd */

if (argc != 2)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string", argv[0]);

if (*argv[1] == 0)
  return Tcl_RetError (interp, "TCL", "Ordinal of null string not defined");

sprintf(interp->result, "%d", (int)(*argv[1]));
return TCL_OK;

} /* Tcl_OrdinalCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ReplicateCmd --
 *     Implements the replicate TCL command:
 *         replicate string count
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns string replicated count times.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_ReplicateCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_ReplicateCmd */
int            RepCount;
register char *SrcPtr;
register char *ScanPtr;
register char *NewPtr;
register int   NewLen;
register int   cnt;

if (argc != 3)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string count", 
                       argv[0]);

if ((!StrToInt (argv[2], 10, &RepCount)) || (RepCount < 0)) 
  return Tcl_RetError (interp, "TCL", "bad count \"%.50s\"", argv[2]);

SrcPtr = argv [1];
NewLen = strlen (SrcPtr) * RepCount;
if (NewLen >= TCL_RESULT_SIZE)
  { /* Must return a dynamic string */
  interp->result = (char *) ckalloc((unsigned) NewLen + 1);
  interp->dynamic = TCL_DYNAMIC;
  } /* Must return a dynamic string */

NewPtr = interp->result;
for (cnt = 0; cnt < RepCount; cnt++)
  { /* Replicate the source string */
  for (ScanPtr = SrcPtr; *ScanPtr != 0; ScanPtr++)
    *NewPtr++ = *ScanPtr;
  } /* Replicate the source string */
*NewPtr = 0;

return TCL_OK;

} /* Tcl_ReplicateCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_StrcatCmd --
 *     Implements the strcat TCL command:
 *         strcat strvar string
 *     See the string(TCL) manual page.
 *
 * Results:
 *     Standard TCL result.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_StrcatCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_StrcatCmd */
char *varPtr;
int   varLen;
int   stringLen;


if (argc != 3)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s strvar string",
                       argv[0]);

varPtr = Tcl_GetVar (interp, argv[1], 0);
if (varPtr == NULL)
  {
  varLen = 0;
  varPtr = "";
  }
else
  varLen = strlen (varPtr);

stringLen = strlen (argv[2]);
if ((varLen + stringLen) >= TCL_RESULT_SIZE)
  {
  interp->result = ckalloc (varLen + stringLen + 1);
  interp->dynamic = 1;
  }

strcpy (interp->result, varPtr);
strcpy (&(interp->result [varLen]), argv[2]);

Tcl_SetVar (interp, argv[1], interp->result, 0);

return TCL_OK;

} /* Tcl_StrcatCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_StrequCmd --
 *     Implements the strequ TCL command:
 *         strequ string1 string2
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Returns 1 if string1 is lexicographically equal to string2.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_StrequCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_StrequCmd */
int result;

if (argc != 3)
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string1 string2",
                       argv[0]);

result = strcmp (argv[1], argv[2]);

sprintf(interp->result, "%d", (result == 0));
return TCL_OK;

} /* Tcl_StrequCmd */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_TrimStringCmd --
 *     Implements the trim_left, trim_right, trim_string TCL commands:
 *         trim_left string [chars]
 *         trim_right string [chars]
 *         trim_string string [chars]
 *     See the string(TCL) manual page.
 *
 * Results:
 *      Return string with all leading whitespaces (spaces, tabs, newlines and
 *      carriage returns) removed.
 *
 *----------------------------------------------------------------------
 */
int
  Tcl_TrimStringCmd (clientData, interp, argc, argv)
    char        *clientData;
    Tcl_Interp  *interp;
    int          argc;
    char       **argv;
{ /* Tcl_TrimStringCmd */
char *trimChars;
char *startPtr;
char *endPtr;
char *checkPtr;
int   len;

if ((argc < 2) || (argc > 3))
  return Tcl_RetError (interp, "TCL", "bad # arg: %.50s string [chars]",
                       argv[0]);

if (argc == 3)
  trimChars = argv[2];
else
  trimChars = " \t\n\r";

/*
 * Trim right and left of string based on command name.
 */

startPtr = argv[1];
endPtr = startPtr + strlen (startPtr) - 1;

if ((argv[0][5] == 'l') || (argv[0][5] == 's'))
  { /* Trim leading characters */

  for (;startPtr != endPtr; startPtr++)
    {
    for (checkPtr = trimChars; *checkPtr != '\0'; checkPtr++)
      if (*checkPtr == *startPtr)
        break;
    if (*checkPtr == '\0')
      break; /* Not in trim characters */
    }

  } /* Trim leading characters */

if ((argv[0][5] == 'r') || (argv[0][5] == 's'))
  { /* Trim trailing characters */

  for (;endPtr >= startPtr; endPtr--)
    {
    for (checkPtr = trimChars; *checkPtr != '\0'; checkPtr++)
      if (*checkPtr == *endPtr)
        break;
    if (*checkPtr == '\0')
      break; /* Not in trim characters */
    }

  } /* Trim trailing characters */

len = (endPtr - startPtr) + 1;
if (len >= TCL_RESULT_SIZE - 1)
  { /* Must return a dynamic string */
  interp->result = (char *) ckalloc((unsigned) len + 1);
  interp->dynamic = TCL_DYNAMIC;
  } /* Must return a dynamic string */

strncpy(interp->result, startPtr, len);
interp->result [len] = '\0';
return TCL_OK;
} /* Tcl_TrimStringCmd */

/*
 *----------------------------------------------------------------------
 *
 *  Tcl_InitString --
 *      Initialize the TCL string extensions.
 *
 *----------------------------------------------------------------------
 */
void
  Tcl_InitString (interp)
    Tcl_Interp *interp;
{
    Tcl_CreateCommand(interp, "char", Tcl_CharCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "cindex", Tcl_CindexCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "clength", Tcl_ClengthCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "crange", Tcl_CrangeCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "substr", Tcl_CrangeCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "lowercase", Tcl_ChangeCaseCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "ordinal", Tcl_OrdinalCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "replicate", Tcl_ReplicateCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "strcat", Tcl_StrcatCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "strequ", Tcl_StrequCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "trim_left", Tcl_TrimStringCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "trim_right", Tcl_TrimStringCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "trim_string", Tcl_TrimStringCmd, 
                     (ClientData)NULL, (void (*)())NULL);
    Tcl_CreateCommand(interp, "uppercase", Tcl_ChangeCaseCmd, 
                     (ClientData)NULL, (void (*)())NULL);
}

