/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "%M% %I% %H%";
#endif

/* Implementation of various unix functions within Tcl */

/* %$*&% Xenix pulls sys/types.h in from within sys/param.h */
#ifndef XENIX286
#  include <sys/types.h>
#endif

#include <sys/param.h>
#include <stdio.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#ifdef BSD
#  define HZ 60
#endif

#include "tcl.h"

extern char **environ;

extern CmdSignal ();
extern SignalTrap ();

/*
 *----------------------------------------------------------------------
 *
 * CmdFork --
 *     Implements the TCL fork command:
 *         fork
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdFork (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int pid;

if (argc != 1)
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s", argv[0]);

pid = fork();
if (pid < 0)
        return Tcl_RetSysError (interp, (char *)NULL);

sprintf(interp->result, "%d", pid);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdWait --
 *     Implements the TCL wait command:
 *         wait
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdWait (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int waitPid;
int status;

if (argc != 1)
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s", argv[0]);

waitPid = wait(&status);
if (waitPid < 0)
        return Tcl_RetSysError (interp, (char *)NULL);

sprintf(interp->result, "%d %d %d", waitPid, status >> 8, status & 0xff);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdKill --
 *     Implements the TCL kill command:
 *        kill [signal] processid
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdKill (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int   signo, procid;
    char *proc_arg;

    if ((argc != 2) && (argc != 3)) 
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s signo procid", 
                             argv[0]);

    if (argc == 3)
        {
        if (!StrToInt (argv[1], 10, &signo))
           if ((signo = SigNameToNum (argv[1])) < 0) 
            return Tcl_RetError (interp, "TCL", 
                                 "Invalid signal name or number");
        }
    else
        signo = 2;
    proc_arg = argv[argc - 1];

    if (!StrToInt (proc_arg, 10, &procid))
            return Tcl_RetError (interp, "TCL", "Invalid process ID");

    if (kill((int)procid,(int)signo) < 0)
        return Tcl_RetSysError (interp, (char *)NULL);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdExit --
 *     Implements the TCL exit command:
 *         exit retcode
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdExit (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
    int status;

    if (argc == 1)
        exit(0);

    if (argc != 2)
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s [status]", 
                             argv[0]);

    if (!StrToInt (argv[1], 10, &status))
            return Tcl_RetError (interp, "TCL", "Invalid exit status");

    exit((int)status);
}

/*
 *----------------------------------------------------------------------
 *
 * CmdChmod --
 *     Implements the TCL chmod command:
 *         chmod mode filename
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdChmod (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int         i, ModeVal;
struct stat FileStat;

if (argc < 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s permbits file [file..]", 
                             argv[0]);

for (i = 2; i < argc; i++)
        { /* Modify each file */

        if ((ModeVal = ConvAbsMode (argv [1])) < 0)
                { /* Try symbolic mode */
                if (stat (argv [i], &FileStat) != 0)
                        return Tcl_RetSysError (interp, (char *)NULL);
                if ((ModeVal = ConvSymMode (argv [1], 
                                 FileStat.st_mode & 07777)) < 0)
                        return Tcl_RetError (interp, "TCL", 
                                             "Invalid file mode");
                } /* Try symbolic mode */

        if (chmod (argv[i], ModeVal) < 0)
                return Tcl_RetSysError (interp, "%s", argv[i]);

        } /* Modify each file */

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdUnlink --
 *     Implements the TCL unlink command:
 *         unlink path [path..]
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdUnlink (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int NameIdx;

if (argc < 2)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s filepath [filepath..]", 
                             argv[0]);

for (NameIdx = 1; NameIdx < argc; NameIdx++)
        if (unlink (argv[NameIdx]) != 0)
                return Tcl_RetSysError (interp, "%s", argv[NameIdx]);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdExecvp --
 *     Implements the TCL execvp command:
 *         execvp prog ["arg1...argN"]
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdExecvp (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
if (argc < 2)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s prog [arg..]", argv[0]);

argv[argc] = NULL;
if (execvp (argv[1], &argv[1]) < 0)
        return Tcl_RetSysError (interp, "%s", argv[1]);

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdGetenv --
 *     Implements the TCL getenv command:
 *         getenv envvar [tclvar]
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdGetenv (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
char *envstr;
extern char *getenv();

if ((argc < 2) || (argc > 3))
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s env_var [tcl_var]", argv[0]);

envstr = getenv(argv[1]);

/* If argc == 3, side-effect assign the third argument as a
 * variable name of the value fetched from the environment
 * variable, and return a 1 if the getenv succeeded or a
 * 0 if it failed.
 * 
 * If only two args were present, return the environment
 * variable value or an empty string if none was found.
 * With the second method it is impossible to differentiate
 * between an empty environment variable and one that was
 * not found.
 */

if (argc == 3)
        {
        if (envstr == NULL)
                {
                Tcl_SetVar(interp, argv[2], "", 0);
                strcpy(interp->result, "0");
                }
        else
                {
                Tcl_SetVar(interp, argv[2], envstr, 0);
                strcpy(interp->result, "1");
                }
        return TCL_OK;
        }

if (envstr != NULL) {       /* Section modified by Glenn M. Lewis - 12/20/90 */
        int size = strlen(envstr);
        if (size >= TCL_RESULT_SIZE) {
            interp->result = ckalloc(size + 1);
            interp->dynamic = TCL_DYNAMIC;
        }
        strcpy(interp->result, envstr);
    }
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdPutenv --
 *     Implements the TCL putenv command:
 *         putenv envvar envvalue
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdPutenv (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
char *EnvStr;

if (argc != 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s envvar envvalue", argv[0]);

EnvStr = ckalloc (strlen (argv[1]) + strlen (argv[2]) + 2);
strcpy (EnvStr, argv[1]);
strcat (EnvStr, "=");
strcat (EnvStr, argv[2]);

if (putenv (EnvStr) < 0)
        panic ("%s: unable to obtain memory to expand environment", argv[0]);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdEnv --
 *     Implements the TCL env command:
 *         env
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdEnv (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
char *result;
int i, nentries, len;
char **strings;

for (nentries = 0; environ[nentries] != NULL; nentries++) ;

strings = (char **)ckalloc(nentries * sizeof(char *));

for (i = 0; i < nentries; i++)
        {
        char *tmpstr;
        int ndx;
        char *pair[2];

        tmpstr = ckalloc((len = strlen(environ[i])) + 1);
        strcpy(tmpstr, environ[i]);
        pair[0] = tmpstr;
        for (ndx = 0; ndx < len; ndx++)
                {
                if (tmpstr[ndx] == '=')
                        {
                        tmpstr[ndx] = '\0';
                        pair[1] = &tmpstr[ndx + 1];
                        goto gotit;
                        }
                }
        return Tcl_RetError (interp, "TCL", 
                             "badly formed string in environment");

        gotit:
        strings[i] = Tcl_Merge(2, pair);
        ckfree(tmpstr);
        }

result = Tcl_Merge(nentries, strings);
for (i = 0; i < nentries; i++)
        ckfree(strings[i]);
ckfree(strings);
Tcl_Return(interp, result, TCL_DYNAMIC);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdTimes --
 *     Implements the TCL times command:
 *         times
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdTimes (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
struct tms tm;

/*
 * Precompute milliseconds-per-tick, the " + HZ / 2" bit gets it to round
 * off instead of truncate.
 */
#define MS_PER_TICK ((1000 + HZ/2) / HZ)

if (argc != 1)
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s", argv[0]);

times(&tm);

sprintf(interp->result, "%ld %ld %ld %ld", 
        tm.tms_utime  * MS_PER_TICK, 
        tm.tms_stime  * MS_PER_TICK, 
        tm.tms_cutime * MS_PER_TICK, 
        tm.tms_cstime * MS_PER_TICK);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdUmask --
 *     Implements the TCL umask command:
 *         umask [octalmask]
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdUmask(clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int mask;

if ((argc < 1) || (argc > 2))
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s octalmask", 
                             argv[0]);

if (argc == 1)
        { /* Return creation mask */

        mask = umask (0);  /* Get current mask      */
        umask (mask);      /* Now set it back (yuk) */
        sprintf (interp->result, "%o", mask);

        } /* Return creation mask */
else
        { /* Set the creation mask */

        if (!StrToInt (argv[1], 8, &mask))
                return Tcl_RetError (interp, "TCL", "bad file creation mask");
        umask(mask);

        } /* Set the creation mask */

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdGetpid --
 *     Implements the TCL getpid command:
 *         getpid
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdGetpid (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
        sprintf (interp->result, "%d", getpid());
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSystem --
 *     Implements the TCL system command:
 *         system command
 *     See the unixcmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL results, may return the UNIX system error message.
 *
 *----------------------------------------------------------------------
 */
int
CmdSystem (clientData, interp, argc, argv)
    char *clientData;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
int exitcode;
extern char *getenv();

if (argc != 2)
        return Tcl_RetError (interp, "TCL", "bad # arg: system command");

exitcode = system(argv[1]);
sprintf(interp->result,"%d",exitcode);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitUnixCmds --
 *      Initializes the TCL unix commands.
 *
 * Side effects:
 *    A catch trap is armmed for the SIGINT signal.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitUnixCmds (interp)
Tcl_Interp *interp;
{
Tcl_CreateCommand(interp, "fork", CmdFork, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "wait", CmdWait, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "kill", CmdKill, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "exit", CmdExit, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "chmod", CmdChmod, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "execvp", CmdExecvp, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "getenv", CmdGetenv, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "env", CmdEnv, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "times", CmdTimes, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "umask", CmdUmask, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "getpid", CmdGetpid, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "system", CmdSystem, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "unlink", CmdUnlink, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand(interp, "putenv", CmdPutenv, (ClientData)NULL,
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "signal", CmdSignal, (ClientData)NULL,
                  (void (*)())NULL);
if (signal(SIGINT, SIG_IGN) != SIG_IGN)
   signal (SIGINT, SignalTrap);
}

